{*******************************************************************************
  Author      : Steve Pinneo (steveyvr@yahoo.ca)
  Updated     : June 10 2003
  Version     : 0.9
  Components  : TTBXImageTabSet

 PURPOSE:

  This component was initually created for the purpose of having a TBX and
  a Windows XP theme aware TTabSet style component. This purpose of this
  component is directed at MDI applications, to SetFocus to the MDIChild forms.

 LICENSE:

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License
  at http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
  the License for the specific language governing rights and
  limitations under the License.

  Alternatively, the contents of this file may be used under the terms
  of the GNU General Public License Version 2 or later (the "GPL"), in
  which case the provisions of the GPL are applicable instead of those
  above. If you wish to allow use of your version of this file only
  under the terms of the GPL and not to allow others to use your
  version of this file under the MPL, indicate your decision by
  deleting the provisions above and replace them with the notice and
  other provisions required by the GPL. If you do not delete the
  provisions above, a recipient may use your version of this file
  under either the MPL or the GPL.

 INSTRUCTIONS:

  1) Set the "Images" property (if you want images on the Tabs).
  2) Add "Items" (Tabs).
  3) Set the individual "Item" properties.

  Optional TBX Theme support:

  4) Add a TTBXSwitcher component to the form, and set the Theme property.
  5) Add the relevant Theme files to the "uses" clause of the same form unit.

 NOTES:

  Improvements and/or Changes to the drawing of the selected Tab may be
  required if you are not satisified with the current appearance. This
  only applies when the component is used on a NON-Windows XP OS, or when
  the Options.EnableXPStyles property is set False.

  Under Windows XP, the Tab is drawn using the current XP TabControl Theme,
  if the Options.EnableXPStyles property is set True (default). The Windows XP
  TabControl theme does not have inverted TabControl support, this applies when
  the "TabPosition" property is set to "tpBottom".

  To override the Windows XP TabControl theme, set the Options.EnableXPStyles
  property to False. The property "BorderColors" does not apply under Windows XP
  when the Options.EnableXPStyles property is set to True.

  This Component can be dockable, but it was not intended to be. :]

*******************************************************************************}

unit TBXImageTabSet;

interface

uses
  Windows, Classes, Graphics, Forms, Controls, Messages, ExtCtrls, ComCtrls,
  ImgList, TB2Dock, TBX;

type
  TTBXImageTabSet = class;

  TTabItem = class(TCollectionItem)
  protected
    FCaption   : string;      //
    FColor     : TColor;      //
    FEnabled   : Boolean;     //
    FHint      : string;      //
    FImageIndex: TImageIndex; //
    FLeft      : Integer;     // Left pixel position of the Tab (internal property)
    FTabIndex  : Integer;     // Used for scoller stop position (internal property)
    FTag       : Integer;     //
    FVisible   : Boolean;     //
    FWidth     : Integer;     // Width in pixels of the Tab (internal property)
    function GetDisplayName: string; override;
    function TabSet: TTBXImageTabSet;
    procedure SetCaption(const Value: string);
    procedure SetColor(const Value: TColor);
    procedure SetEnabled(const Value: Boolean);
    procedure SetHint(const Value: string);
    procedure SetImageIndex(Value: TImageIndex);
    procedure SetTag(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
    property Left    : Integer read FLeft     write FLeft default -1;
    property TabIndex: Integer read FTabIndex write FTabIndex default -1;
    property Width   : Integer read FWidth    write FWidth default -1;
  published
    property Caption   : string      read FCaption    write SetCaption;
    property Color     : TColor      read FColor      write SetColor default clNone;
    property Enabled   : Boolean     read FEnabled    write SetEnabled default True;
    property Hint      : string      read FHint       write SetHint;
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default -1;
    property Tag       : Integer     read FTag        write SetTag default 0;
    property Visible   : Boolean     read FVisible    write SetVisible default True;
  end;

  TTabCollection = class(TCollection)
  private
    FTabSet: TTBXImageTabSet;
    function GetItem(Index: Integer): TTabItem;
    procedure SetItem(Index: Integer; Value: TTabItem);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(TabSet: TTBXImageTabSet);
    procedure Assign(Source: TPersistent); override;
    function Add: TTabItem;
    procedure Clear;
    procedure ClearTabInfo; // Sets the Items:  Left, TabIndex, Width's to  -1.
    procedure Delete(Index: Integer);
    // NOTE: DeleteByCaption and DeleteByHint come in handy when the Tab control
    //       is being used to hold Filenames, or something similar.
    //       As an example... in an MDI application, the Tabs can represent open
    //       files. The developer can store the ExtractFileName(filename) in the
    //       Caption property, and store the path+filename in the Hint property.
    procedure DeleteByCaption(ACaption: string);
    procedure DeleteByHint(AHint: string);
    function  Insert(Index: Integer): TTabItem;
    property Items[Index: Integer]: TTabItem read GetItem write SetItem; default;
    property TabSet: TTBXImageTabSet read FTabSet;
  end;

  TTabColors = class(TPersistent)
  private
    FHighlight    : TColor;
    FHighlightDark: TColor;
    FShadow       : TColor;
    FShadowDark   : TColor;
    FOnChange     : TNotifyEvent;
    procedure SetHighlight(const Value: TColor);
    procedure SetHighlightDark(const Value: TColor);
    procedure SetShadow(const Value: TColor);
    procedure SetShadowDark(const Value: TColor);
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    constructor Create;
    procedure Changed; virtual;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property Highlight    : TColor read FHighlight     write SetHighlight default clBtnHighlight;
    property HighlightDark: TColor read FHighlightDark write SetHighlightDark default cl3DLight;
    property Shadow       : TColor read FShadow        write SetShadow default clBtnShadow;
    property ShadowDark   : TColor read FShadowDark    write SetShadowDark default cl3DDkShadow;
  end;

  TTabOptions = class(TPersistent)
  private
    FChangeOnScroll: Boolean;
    FColors        : TTabColors;
    FEnableXPStyles: Boolean;
    FLeftMargin    : Integer;
    FRightMargin   : Integer;
    FTabMargin     : Integer;
    FOnChange      : TNotifyEvent;
    procedure SetChangeOnScroll(const Value: Boolean);
    procedure SetEnableXPStyles(const Value: Boolean);
    procedure SetLeftMargin(const Value: Integer);
    procedure SetRightMargin(const Value: Integer);
    procedure SetTabMargin(const Value: Integer);
  protected
    procedure AssignTo(Dest: TPersistent); override;
    procedure ColorsChanged(Sender: TObject);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Changed; virtual;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  published
    property BorderColors  : TTabColors read FColors         write FColors;
    property ChangeOnScroll: Boolean    read FChangeOnScroll write SetChangeOnScroll default False;
    property EnableXPStyles: Boolean    read FEnableXPStyles write SetEnableXPStyles default True;
    property LeftMargin    : Integer    read FLeftMargin     write SetLeftMargin default 0;
    property RightMargin   : Integer    read FRightMargin    write SetRightMargin default 1;
    property TabMargin     : Integer    read FTabMargin      write SetTabMargin default 9;    
  end;

  TTabScroller = class(TCustomUpDown)
  private
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
  end;

  TChangeTabEvent   = procedure(Sender: TObject; Index: Integer) of object;
  TChangingTabEvent = procedure(Sender: TObject; Index: Integer; var AllowChange: Boolean) of object;
  TGetTabImageEvent = procedure(Sender: TObject; Index: Integer; var ImageIndex: Integer) of object;
  TMeasureTabEvent  = procedure(Sender: TObject; Index: Integer; var TabWidth: Integer) of object;

  TTabPosition = (tpBottom, tpTop);

  TTBXImageTabSet = class(TCustomControl)
  private
    FEffectiveColor : TColor;
    FFirstIndex     : Integer;
    FImageChangeLink: TChangeLink;
    FImages         : TCustomImageList;
    FItems          : TTabCollection;
    FMouseOverTab   : Integer;
    FNeedAdjust     : Boolean;
    FOptions        : TTabOptions;
    FPosition       : TTabPosition;
    FScrollBtn      : TTabScroller;
    FTabHeight      : Integer;
    FTabIndex       : Integer;
    FVisibleTabs    :  Integer;
    // Events
    FOnChanging     : TChangingTabEvent;
    FOnChange       : TChangeTabEvent;
    FOnGetImage     : TGetTabImageEvent;
    FOnMeasureTab   : TMeasureTabEvent;
    FOnScrollClick  : TUDClickEvent;
    procedure SetFirstIndex(const Value: Integer);
    procedure SetImages(Value: TCustomImageList);
    procedure SetTabPosition(const Value: TTabPosition);
    procedure SetTabHeight(const Value: Integer);
    procedure SetTabIndex(const Value: Integer);
    function GetTabItems: TTabCollection;
    procedure SetTabItems(Value: TTabCollection);
    //
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMDialogChar(var Message: TCMDialogChar); message CM_DIALOGCHAR;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMParentColorChanged(var Message: TMessage); message CM_PARENTCOLORCHANGED;
    procedure TBMGetEffectiveColor(var Message: TMessage); message TBM_GETEFFECTIVECOLOR;
    procedure TBMThemeChange(var Message: TMessage); message TBM_THEMECHANGE;
    procedure WMEraseBkgnd(var Message: TWmEraseBkgnd); message WM_ERASEBKGND;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMNCHitTest(var message: TWMNCHitTest); message WM_NCHITTEST;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure AdjustTabPos;
    procedure AdjustScrollBtnPos;
    function CalcTabPositions(Start, Stop: Integer; Canvas: TCanvas): Integer;
    procedure ImageListChange(Sender: TObject);
    procedure ReadBoolData(Reader: TReader);
    procedure ReadIntData(Reader: TReader);
    procedure ScrollButtonClick(Sender: TObject; Button: TUDBtnType);
  protected
    function CanChange(NewIndex: Integer): Boolean;
    procedure Change; dynamic;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure DefineProperties(Filer: TFiler); override;
    procedure DisplayTabHint(TabIndex: Integer);
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
    function GetImageIndex(Index: Integer): Integer; virtual;
    procedure MeasureTab(Index: Integer; var TabWidth: Integer); virtual;
    procedure OptionsChange(Sender: TObject);
    procedure Paint; override;
    procedure UpdateEffectiveColor;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property EffectiveColor: TColor read FEffectiveColor;
    function HasTabItems: Boolean;
    function ItemAtPos(Pos: TPoint): Integer;
    function ItemRect(Item: Integer): TRect;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    //
    property Canvas;
    property FirstIndex: Integer read FFirstIndex write SetFirstIndex default 0;
    property VisibleTabs: Integer read FVisibleTabs;
  published
    property Images     : TCustomImageList read FImages     write SetImages;
    property Items      : TTabCollection   read GetTabItems write SetTabItems stored HasTabItems;
    property Options    : TTabOptions      read FOptions    write FOptions;
    property TabHeight  : Integer          read FTabHeight  write SetTabHeight default 22;
    property TabIndex   : Integer          read FTabIndex   write SetTabIndex default -1;
    property TabPosition: TTabPosition     read FPosition   write SetTabPosition default tpTop;
    // Events
    property OnChanging     : TChangingTabEvent read FOnChanging    write FOnChanging;
    property OnChange       : TChangeTabEvent   read FOnChange      write FOnChange;
    property OnGetImageIndex: TGetTabImageEvent read FOnGetImage    write FOnGetImage;
    property OnMeasureTab   : TMeasureTabEvent  read FOnMeasureTab  write FOnMeasureTab;
    property OnScrollerClick: TUDClickEvent     read FOnScrollClick write FOnScrollClick;
    // Inherited properties and events
    property Align;
    property Anchors;
    property Color default clNone;
    property Constraints;
    property Cursor;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Height;
    property HelpContext;
    property HelpKeyword;
    property HelpType;        
    property Hint;
    property Left;        
    property Name;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Tag;
    property Top;
    property Width;
    property Visible;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

implementation

uses
  Consts, SysUtils, TBXThemesEx, TBXThemes, TBXUxThemes;

const
  MINIMUM_CONTROL_HEIGHT: Integer = 12;

{ TTabItem }

constructor TTabItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FCaption    := EmptyStr;
  FColor      := clNone;
  FEnabled    := True;
  FHint       := EmptyStr;
  FImageIndex := -1;
  FLeft       := -1;
  FTabIndex   := -1;
  FTag        := 0;
  FVisible    := True;
  FWidth      := -1;
end;

procedure TTabItem.Assign(Source: TPersistent);
begin
  if (Source is TTabItem) then
  begin
    Collection.BeginUpdate;
    try
      FCaption    := TTabItem(Source).Caption;
      FColor      := TTabItem(Source).Color;
      FEnabled    := TTabItem(Source).Enabled;
      FHint       := TTabItem(Source).Hint;
      FImageIndex := TTabItem(Source).ImageIndex;
      FTag        := TTabItem(Source).Tag;
      FVisible    := TTabItem(Source).Visible;
      FLeft       := -1;
      FTabIndex   := -1;
      FWidth      := -1;
    finally
      Collection.EndUpdate;
    end;
  end
  else
    inherited Assign(Source);
end;

function TTabItem.GetDisplayName: string;
begin
  if Length(FCaption) > 0 then
    Result := FCaption
  else
    Result := inherited GetDisplayName;
end;

function TTabItem.TabSet: TTBXImageTabSet;
begin
  Result := TTabCollection(Collection).TabSet;
end;

procedure TTabItem.SetCaption(const Value: string);
begin
  if (FCaption <> Value) then
  begin
    FCaption := Value;
    Changed(False);
  end;
end;

procedure TTabItem.SetColor(const Value: TColor);
begin
  if (FColor <> Value) then
  begin
    FColor := Value;
    Changed(False);
  end;
end;

procedure TTabItem.SetEnabled(const Value: Boolean);
begin
  if (FEnabled <> Value) then
  begin
    FEnabled := Value;
    Changed(False);
  end;
end;

procedure TTabItem.SetHint(const Value: string);
begin
  if (FHint <> Value) then
  begin
    FHint := Value;
    Changed(False);
  end;
end;

procedure TTabItem.SetImageIndex(Value: TImageIndex);
begin
  if (FImageIndex <> Value) then
  begin
    FImageIndex := Value;
    if Assigned(TTBXImageTabSet(Collection).Images) then
      Changed(False);
  end;
end;

procedure TTabItem.SetTag(const Value: Integer);
begin
  if (FTag <> Value) then
  begin
    FTag := Value;
    Changed(False);
  end;
end;

procedure TTabItem.SetVisible(const Value: Boolean);
begin
  if (FVisible <> Value) then
  begin
    FVisible := Value;
    Changed(False);
  end;
end;

{ TTabCollection }

constructor TTabCollection.Create(TabSet: TTBXImageTabSet);
begin
  inherited Create(TTabItem);
  FTabSet := TabSet;
end;

function TTabCollection.GetOwner: TPersistent;
begin
  Result := FTabSet;
end;

procedure TTabCollection.Update(Item: TCollectionItem);
  //
  function LocateNextEnabledVisible(StartIndex: Integer): Integer;
  var
    i: Integer;
  begin
    Result := -1;
    for i := StartIndex to Count-1 do
    begin
      if Items[i].Enabled and Items[i].Visible then
      begin
        Result := i;
        Break;
      end;
    end;
    if (Result = -1) then
    begin
      for i := 0 to StartIndex do
      begin
        if Items[i].Enabled and Items[i].Visible then
        begin
          Result := i;
          Break;
        end;
      end;
    end;
  end;
begin
  if Assigned(FTabSet) then
  begin
    if (Count > 0) and (FTabSet.TabIndex = -1) then
      FTabSet.TabIndex := 0;

    if (Count > 0) and (FTabSet.TabIndex in [0..Count-1]) then
    begin
      if not (Items[FTabSet.TabIndex].Enabled and Items[FTabSet.TabIndex].Visible) then
        FTabSet.TabIndex := LocateNextEnabledVisible(FTabSet.TabIndex);
    end;

    FTabSet.Invalidate;
  end;
end;

procedure TTabCollection.Assign(Source: TPersistent);
var
  i: Integer;
begin
  if (Source is TTabCollection) then
  begin
    BeginUpdate;
    try
      Clear;
      for i := 0 to TTabCollection(Source).Count-1 do
        Add.Assign(TTabCollection(Source).Items[i]);
    finally
      EndUpdate;
    end;
  end
  else
    inherited Assign(Source);
end;

function TTabCollection.Add: TTabItem;
begin
  Result := TTabItem(inherited Add);
end;

procedure TTabCollection.Clear;
begin
  inherited Clear;
  FTabSet.TabIndex := -1;
end;

procedure TTabCollection.ClearTabInfo;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
  begin
    TTabItem(Items[i]).Left     := -1;
    TTabItem(Items[i]).TabIndex := -1;
    TTabItem(Items[i]).Width    := -1;
  end;
end;

procedure TTabCollection.Delete(Index: Integer);
var
  iTabIndex: Integer;
begin
  if (Index in [0..Count-1]) then
  begin
    iTabIndex := FTabSet.TabIndex;
    inherited Delete(Index);

    if (iTabIndex > Index) and (Count > 0) then
      FTabSet.TabIndex := iTabIndex - 1
    else
    if (Count = 0) then
      FTabSet.TabIndex := -1;
  end;
end;

procedure TTabCollection.DeleteByCaption(ACaption: string);
var
  i: Integer;
begin
  for i := 0 to Count-1 do
  begin
    if (Items[i].Caption = ACaption) then
    begin
      Delete(i);
      Break;
    end;
  end;
end;

procedure TTabCollection.DeleteByHint(AHint: string);
var
  i: Integer;
begin
  for i := 0 to Count-1 do
  begin
    if (Items[i].Hint = AHint) then
    begin
      Delete(i);
      Break;
    end;
  end;
end;

function TTabCollection.Insert(Index: Integer): TTabItem;
var
  iTabIndex: Integer;
begin
  if (Index > -1) then
  begin
    iTabIndex := FTabSet.TabIndex;
    Result := TTabItem(inherited Insert(Index));
    if (Result <> nil) then
    begin
      if (iTabIndex >= Index) and (Count > 1) then
        FTabSet.TabIndex := iTabIndex + 1;
    end;
  end
  else
    Result := nil;
end;

function TTabCollection.GetItem(Index: Integer): TTabItem;
begin
  Result := TTabItem(inherited GetItem(Index));
end;

procedure TTabCollection.SetItem(Index: Integer; Value: TTabItem);
begin
  inherited SetItem(Index, Value);
end;

{ TTabColors }

constructor TTabColors.Create;
begin
  inherited Create;
  FHighlight     := clBtnHighlight;
  FHighlightDark := cl3DLight;
  FShadow        := clBtnShadow;
  FShadowDark    := cl3DDkShadow;
end;

procedure TTabColors.AssignTo(Dest: TPersistent);
begin
  if (Dest is TTabColors) then
  begin
    with TTabColors(Dest) do
    begin
      FHighlight     := Self.Highlight;
      FHighlightDark := Self.HighlightDark;
      FShadow        := Self.Shadow;
      FShadowDark    := Self.ShadowDark;
      Changed;
    end
  end
  else
    inherited AssignTo(Dest);
end;

procedure TTabColors.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TTabColors.SetHighlight(const Value: TColor);
begin
  if (FHighlight <> Value) then
  begin
    FHighlight := Value;
    Changed;
  end;
end;

procedure TTabColors.SetHighlightDark(const Value: TColor);
begin
  if (FHighlightDark <> Value) then
  begin
    FHighlightDark := Value;
    Changed;
  end;
end;

procedure TTabColors.SetShadow(const Value: TColor);
begin
  if (FShadow <> Value) then
  begin
    FShadow := Value;
    Changed;
  end;
end;

procedure TTabColors.SetShadowDark(const Value: TColor);
begin
  if (FShadowDark <> Value) then
  begin
    FShadowDark := Value;
    Changed;
  end;
end;

{ TTabOptions }

constructor TTabOptions.Create;
begin
  inherited Create;
  FChangeOnScroll := False;
  FColors         := TTabColors.Create;
  FEnableXPStyles := True;
  FLeftMargin     := 0;
  FRightMargin    := 1;
  FTabMargin      := 9;
  FColors.OnChange  := ColorsChanged;
end;

destructor TTabOptions.Destroy;
begin
  FColors.Free;
  inherited Destroy;
end;

procedure TTabOptions.AssignTo(Dest: TPersistent);
begin
  if (Dest is TTabOptions) then
  begin
    with TTabOptions(Dest) do
    begin
      FColors         := Self.BorderColors;
      FChangeOnScroll := Self.ChangeOnScroll;
      FEnableXPStyles := Self.EnableXPStyles;
      FLeftMargin     := Self.LeftMargin;
      FRightMargin    := Self.RightMargin;
      FTabMargin      := Self.TabMargin;      
      Changed;
    end
  end
  else
    inherited AssignTo(Dest);
end;

procedure TTabOptions.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TTabOptions.ColorsChanged(Sender: TObject);
begin
  Changed;
end;

procedure TTabOptions.SetChangeOnScroll(const Value: Boolean);
begin
  if (FChangeOnScroll <> Value) then
  begin
    FChangeOnScroll := Value;
    Changed;
  end;
end;

procedure TTabOptions.SetEnableXPStyles(const Value: Boolean);
begin
  if (FEnableXPStyles <> Value) then
  begin
    FEnableXPStyles := Value;
    Changed;
  end;
end;

procedure TTabOptions.SetLeftMargin(const Value: Integer);
begin
  if (FLeftMargin <> Value)  then
  begin
    FLeftMargin := Value;
    Changed;
  end;
end;

procedure TTabOptions.SetRightMargin(const Value: Integer);
begin
  if (FRightMargin <> Value) then
  begin
    FRightMargin := Value;
    Changed;
  end;
end;

procedure TTabOptions.SetTabMargin(const Value: Integer);
begin
  if (FTabMargin <> Value) then
  begin
    FTabMargin := Value;
    Changed;
  end;
end;

{ TTabScroller }

procedure TTabScroller.WMEraseBkgnd(var Message: TWmEraseBkgnd);
begin
  Message.Result := 1;
end;

{ TTBXImageTabSet }

constructor TTBXImageTabSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := [csCaptureMouse, csClickEvents, csDoubleClicks];
  if not (csDesigning in ComponentState) then
    ControlStyle := ControlStyle - [csOpaque];
  DoubleBuffered := True;
  Width  := 300;
  Height := 23;
  ParentColor := False;
  ParentFont  := True;

  FEffectiveColor := clNone;
  Color := clNone;

  FFirstIndex   := 0;
  FMouseOverTab := -1;
  FTabHeight    := 22;
  FPosition     := tpTop;
  FTabIndex     := -1;
  FVisibleTabs  := 0;

  FImageChangeLink := TChangeLink.Create;
  FImageChangeLink.OnChange := ImageListChange;

  FItems := TTabCollection.Create(Self);

  FOptions := TTabOptions.Create;
  FOptions.OnChange := OptionsChange;

  FScrollBtn := TTabScroller.Create(Self);
  with FScrollBtn do
  begin
    Parent      := Self;         // The Parent is resposible for Destruction.
    Orientation := udHorizontal; // This must be set prior to Width and Height.
    Visible     := False;
    Height      := 17;
    Width       := 34;
    Top         := 0;
    Min         := 0;
    Max         := 0;
    Position    := 0;
    OnClick     := ScrollButtonClick;
  end;

  AddThemeNotification(Self);
end;

procedure TTBXImageTabSet.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params.WindowClass do
    Style := Style and not (CS_VREDRAW or CS_HREDRAW);
end;

destructor TTBXImageTabSet.Destroy;
begin
  RemoveThemeNotification(Self);
  FImageChangeLink.Free;
  FItems.Free;
  FOptions.Free;
  inherited Destroy;
end;

procedure TTBXImageTabSet.OptionsChange(Sender: TObject);
begin
  Invalidate;
end;

procedure TTBXImageTabSet.AdjustScrollBtnPos;
  // Math
  function Max(const A, B: Integer): Integer;
  begin
    if A > B then Result := A else Result := B;
  end;
  function Min(const A, B: Integer): Integer;
  begin
    if A < B then Result := A else Result := B;
  end;
begin
  if Assigned(FScrollBtn) then
  begin
    FScrollBtn.Left   := Width - FScrollBtn.Width;
    FScrollBtn.Height := Min(Max(Height-5, 10), 17);
  end;
end;

procedure TTBXImageTabSet.AdjustTabPos;
var
  iLastVisibleTab: Integer;
  //
  function GetRightSide: Integer;
  begin
    Result := Width - Options.RightMargin;
    if (FVisibleTabs < Items.Count-1) then
      Dec(Result, FScrollBtn.Width + 2);
  end;
  //
  function ReverseCalcNumTabs(Start, Stop: Integer; Canvas: TCanvas): Integer;
  var
    iWidth: Integer;
  begin
    if HandleAllocated then
    begin
      Result := FTabIndex;
      while (Start >= Stop) and (Result >= 0) do
        with Canvas do
        begin
          iWidth := TextWidth(Items[Result].Caption);
          MeasureTab(Result, iWidth);
          Dec(Start, iWidth + Options.TabMargin); // Next usable position
          if (Start >= Stop) then
            Dec(Result);
        end;
      if (Start < Stop) or (Result < 0) then
        Inc(Result);
    end
    else
      Result := FFirstIndex;
  end;
begin
  if (Items.Count > 0) then
  begin
    iLastVisibleTab := FFirstIndex + FVisibleTabs - 1;
    if (FTabIndex > iLastVisibleTab) then
      FFirstIndex := ReverseCalcNumTabs(GetRightSide,
        Options.LeftMargin + Options.TabMargin, Canvas)
    else
    if (FTabIndex > -1) and (FTabIndex < FFirstIndex) then
      FFirstIndex := FTabIndex;
  end;
end;

function TTBXImageTabSet.CalcTabPositions(Start, Stop: Integer; Canvas: TCanvas): Integer;
type
  TTabInfo = record
    TabIndex: Integer;
    Left    : Integer;
    Width   : Integer;
  end;
var
  iIndex : Integer;
  iWidth : Integer;
  TabInfo: TTabInfo;
begin
  Items.ClearTabInfo;

  iIndex := FirstIndex;
  while (Start < Stop) and (iIndex < Items.Count) do
  begin
    with Canvas do
    begin
      if Items[iIndex].Visible then
      begin
        TabInfo.TabIndex := iIndex;
        TabInfo.Left := Start;
        iWidth := TextWidth(Items[iIndex].Caption) + 2;
        MeasureTab(iIndex, iWidth);
        if Assigned(FImages) then
          TabInfo.Width := iWidth + FImages.Width + 4
        else
          TabInfo.Width := iWidth;
        Inc(Start, TabInfo.Width + Options.TabMargin); // Next usable position
      end;

      if (Start <= Stop) then
      begin
        Items[iIndex].Left     := TabInfo.Left;
        Items[iIndex].TabIndex := TabInfo.TabIndex;
        Items[iIndex].Width    := TabInfo.Width;
        Inc(iIndex);
      end;
    end;
  end;
  Result := iIndex - FirstIndex;
end;

function TTBXImageTabSet.CanChange(NewIndex: Integer): Boolean;
begin
  Result := False;

  if (NewIndex in [0..Items.Count-1]) then
  begin
    if Items[NewIndex].Enabled and Items[NewIndex].Visible then
    begin
      Result := True;
      if not (csDesigning in ComponentState) then
        if Assigned(FOnChanging) then
          FOnChanging(Self, NewIndex, Result);
    end;
  end;
end;

procedure TTBXImageTabSet.Change;
var
  Form: TCustomForm;
begin
  if (csDesigning in ComponentState) then
  begin
    Form := GetParentForm(Self);
    if (Form <> nil) and (Form.Designer <> nil) then
      Form.Designer.Modified;
  end
  else
    if Assigned(FOnChange) then
      FOnChange(Self, TabIndex);
end;

procedure TTBXImageTabSet.DisplayTabHint(TabIndex: Integer);
var
  sHint: string;
begin
  if (TabIndex in [0..Items.Count-1]) then
  begin
    sHint := Items[TabIndex].Hint;
    if (sHint <> EmptyStr) then
    begin
      Application.CancelHint;
      Hint := sHint;
    end;
  end;
end;

procedure TTBXImageTabSet.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
  //
end;

function TTBXImageTabSet.GetImageIndex(Index: Integer): Integer;
begin
  if (Index in [0..Items.Count-1]) then
    Result := Items[Index].ImageIndex
  else
    Result := -1;

  if not (csDesigning in ComponentState) then
    if Assigned(FOnGetImage) then
      FOnGetImage(Self, Index, Result);
end;

function TTBXImageTabSet.HasTabItems: Boolean;
begin
  Result := (Items <> nil) and (Items.Count > 0);
end;

procedure TTBXImageTabSet.ImageListChange(Sender: TObject);
begin
  if (Sender = Images) then
    Invalidate;
end;

function TTBXImageTabSet.ItemAtPos(Pos: TPoint): Integer;
var
  i: Integer;
begin
  Result := -1;
  if not (Pos.Y in [0..ClientHeight]) then Exit;

  for i := 0 to Items.Count-1 do
  begin
    if (Items[i].Width = -1) or (not Items[i].Enabled) then
      Continue;

    if (Pos.X - Items[i].Left in [0..Items[i].Width]) then
    begin
      if (FPosition = tpBottom) and (Pos.Y in [0..FTabHeight]) then
      begin
        Result := FScrollBtn.Position + i;
        Exit;
      end
      else
      if (Pos.Y in [ClientHeight - FTabHeight..FTabHeight]) then
      begin
        Result := FScrollBtn.Position + i;
        Exit;
      end;
    end;
  end;
end;

function TTBXImageTabSet.ItemRect(Item: Integer): TRect;
begin
  if (Item in [0..Items.Count-1]) then
    Result := Rect(Items[Item].Left, 0, Items[Item].Left + Items[Item].Width, FTabHeight)
  else
    Result := Rect(0, 0, 0, 0);
end;

procedure TTBXImageTabSet.MeasureTab(Index: Integer; var TabWidth: Integer);
begin
  if not (csDesigning in ComponentState) then
    if Assigned(FOnMeasureTab) then
      FOnMeasureTab(Self, Index, TabWidth);
end;

procedure TTBXImageTabSet.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = Images) then
    Images := nil;
end;

procedure TTBXImageTabSet.Paint;
  // Draw the control Background.
  procedure DrawBackground(ACanvas: TCanvas; ARect: TRect);
  begin
    CurrentTheme.PaintBackgnd(ACanvas, ARect, ARect, ARect, EffectiveColor,
        False, TVT_NORMALTOOLBAR);
  end;
  // Draw the tab Background.
  procedure DrawTabBackground(ACanvas: TCanvas; ARect: TRect);
  begin
    if USE_THEMES and Options.EnableXPStyles then
      //CurrentTheme.PaintTabSheet(ACanvas.Handle, ARect, TABP_TABITEM, TIS_SELECTED)
      DrawThemeBackground(TAB_THEME, ACanvas.Handle, TABP_TABITEM, TIS_SELECTED, ARect, nil)
    else
    begin
      ACanvas.Brush.Color := EffectiveColor;
      ACanvas.FillRect(ARect);
    end;
  end;
  // Get and Draw the tab Image.
  function DrawTabImage(ACanvas: TCanvas; var ARect: TRect; ATabIndex: Integer;
      ASelected, AEnabled: Boolean): Boolean;
  var
    iImgIndex: Integer;
    ItemInfo : TTBXItemInfo;
  begin
    Result := False;
    if Assigned(FImages) then
    begin
      iImgIndex := GetImageIndex(ATabIndex);
      if (iImgIndex in [0..FImages.Count-1]) then
      begin
        Result := True;
        FillChar(ItemInfo, SizeOf(TTBXItemInfo), 0);
        with ItemInfo do
        begin
          ViewType    := TVT_MENUBAR;
          ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE or IO_DESIGNING;
          Enabled     := Items[ATabIndex].Enabled;
          Pushed      := False;
          Selected    := ASelected;
          ImageShown  := True;
          ImageHeight := FImages.Height;
          ImageWidth  := FImages.Width;
          IsVertical  := False;
        end;
        CurrentTheme.PaintImage(ACanvas, ARect, ItemInfo, FImages, iImgIndex);
      end;
    end;
  end;
  // Draw the tab Caption.
  procedure DrawTabCaption(ACanvas: TCanvas; var ARect: TRect; ATabIndex: Integer;
      ASelected, AEnabled, DoCalc: Boolean);
  var
    ItemInfo: TTBXItemInfo;
    Flags: Cardinal;
  begin
    Inc(ARect.Left, 2);
    FillChar(ItemInfo, SizeOf(TTBXItemInfo), 0);
    with ItemInfo do
    begin
      ViewType    := TVT_MENUBAR;
      ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE or IO_DESIGNING;
      Enabled     := Items[ATabIndex].Enabled;
      Pushed      := False;
      Selected    := ASelected;
      ImageShown  := False;
      ImageHeight := 0;
      ImageWidth  := 0;
      IsVertical  := False;
    end;
    Flags := DT_CENTER or DT_VCENTER or DT_SINGLELINE;
    if DoCalc then
      Flags := Flags or DT_CALCRECT;
    CurrentTheme.PaintCaption(ACanvas, ARect, ItemInfo, Items[ATabIndex].Caption, Flags, False);
  end;
  // Draw the tab Separator's.
  procedure DrawTabSeparator(ACanvas: TCanvas; ARect: TRect);
  var
    ItemInfo: TTBXItemInfo;
  begin
    FillChar(ItemInfo, SizeOf(TTBXItemInfo), 0);
    with ItemInfo do
    begin
      ViewType    := TVT_MENUBAR;
      ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE or IO_DESIGNING;
      Enabled     := True;
      Pushed      := False;
      Selected    := False;
      ImageShown  := False;
      ImageWidth  := 0;
      ImageHeight := 0;
      IsVertical  := True;
    end;
    CurrentTheme.PaintSeparator(ACanvas, ARect, ItemInfo, False, True);
  end;
var
  i: Integer;
  rec: TRect;
  recSep: TRect;
  oMemBmp: TBitmap;
  oSepBmp: TBitmap;
  oImages: TCustomImageList;
  iHalfTabMargin: Integer;
  iClientRight  : Integer;
  iClientLeft   : Integer;
  bIsFirst      : Boolean;
  bIsFirstTab   : Boolean;
  bIsLastTab    : Boolean;
  bHasImage     : Boolean;  
  TabItem       : TTabItem;
  ItemInfo      : TTBXItemInfo;
begin
  if not HandleAllocated then Exit;

  AdjustScrollBtnPos;
  iHalfTabMargin := Options.TabMargin div 2;

  oMemBmp := TBitmap.Create;
  oSepBmp := TBitmap.Create;
  try
    oMemBmp.Width  := ClientWidth;
    oMemBmp.Height := ClientHeight;
    oMemBmp.Canvas.Font := Self.Canvas.Font;

    oSepBmp.Width  := 4;
    oSepBmp.Height := ClientHeight;

    // Set the Tabs starting at this position.
    iClientLeft := Options.TabMargin + Options.LeftMargin;
    // Draw the Tabs up to this position.
    iClientRight := Width - Options.RightMargin;

    // Calculate the tab position info (ALL of the Tabs).
    FVisibleTabs := CalcTabPositions(iClientLeft, iClientRight, oMemBmp.Canvas);

    // Set the scroller.Max property (*NOT* ALL of the tabs are visible).
    if (VisibleTabs < Items.Count) then
    begin
      if Options.ChangeOnScroll then
        // Display ALL the tabs (Scrollable).
        FScrollBtn.Max := Items.Count - 1
      else
      begin
        // Display ONLY the tabs that will fit according to the control Width.
        Dec(iClientRight, FScrollBtn.Width + 2);
        FVisibleTabs := CalcTabPositions(iClientLeft, iClientRight, oMemBmp.Canvas);
        FScrollBtn.Max := Items.Count - VisibleTabs;
      end;
    end;

    // Show or Hide the scroller.
    if (VisibleTabs < Items.Count) then
    begin
      // All of the tabs are *NOT* visible, show the Scroller.
      FScrollBtn.Min := 0;
      FScrollBtn.Position := FirstIndex;
      FScrollBtn.Visible := True;
      ShowWindow(FScrollBtn.Handle, SW_SHOW);
    end
    else
    begin
      // ALL of the tabs are visible, hide the Scroller.
      FScrollBtn.Visible := False;
      ShowWindow(FScrollBtn.Handle, SW_HIDE);
    end;

    // The control has been Resized. :[
    if FNeedAdjust then
    begin
      AdjustTabPos;
      FVisibleTabs := CalcTabPositions(iClientLeft, iClientRight, oMemBmp.Canvas);
    end;
    FNeedAdjust := False;

    with oMemBmp.Canvas do
    begin
      // Draw the control Background.
      if (FPosition = tpBottom) then
        rec := Rect(0, 0, oMemBmp.Width, oMemBmp.Height)
      else
        rec := Rect(0, 0, oMemBmp.Width, oMemBmp.Height);
      DrawBackground(oMemBmp.Canvas, rec);

      if (csDesigning in ComponentState) then
      begin
        Pen.Style   := psDot;
        Pen.Color   := clBtnShadow;
        Brush.Style := bsClear;
        Rectangle(rec.Left, rec.Top, rec.Right, rec.Bottom);
        Pen.Style := psSolid;
      end;
    end;

    // Draw the Tab's.
    for i := FirstIndex to Items.Count-1 do
    begin
      TabItem := Items[i];

      if (not TabItem.Visible) or (TabItem.Width = -1) then
        Continue;

      bIsFirst    := (i = 0);
      bIsFirstTab := bIsFirst and (FScrollBtn.Position = 0);

      with oMemBmp.Canvas do
      begin
        ////////////////////////////////////////////////////////////////////////
        // Draw the Selected Tab ///////////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////
        if (i = TabIndex) then
        begin
          // Draw the tab Background.
          if USE_THEMES and Options.EnableXPStyles then
          begin
            if (FPosition = tpBottom) then
              // Tabs are on the Bottom.
              rec := Rect(TabItem.Left - iHalfTabMargin, 0,
                          TabItem.Left + iHalfTabMargin + TabItem.Width + 1, FTabHeight)
            else
              // Tabs are on the Top.
              rec := Rect(TabItem.Left - iHalfTabMargin, Height - FTabHeight,
                          TabItem.Left + iHalfTabMargin + TabItem.Width + 1, Height + 1);
          end
          else
          begin
            if (FPosition = tpBottom) then
              // Tabs are on the Bottom, (Top = -1) cuz of the Button border.
              rec := Rect(TabItem.Left - iHalfTabMargin, -1,
                          TabItem.Left + iHalfTabMargin + TabItem.Width + 1, FTabHeight)
            else
              // Tabs are on the Top, (Bottom = Height + 1) cuz of the Button border.
              rec := Rect(TabItem.Left - iHalfTabMargin, Height - FTabHeight + 1,
                          TabItem.Left + iHalfTabMargin + TabItem.Width + 1, Height + 1);
          end;
          DrawTabBackground(oMemBmp.Canvas, rec);

          // Draw the tab Image.
          Brush.Style := bsClear;
          if (FPosition = tpBottom) then
            rec := Rect(TabItem.Left, 2,
                        TabItem.Left + TabItem.Width, FTabHeight)
          else
            rec := Rect(TabItem.Left, Height - FTabHeight + 2,
                        TabItem.Left + TabItem.Width, Height);
          bHasImage := DrawTabImage(oMemBmp.Canvas, rec, i, True, True);

          // Draw the tab Caption.
          Brush.Style := bsClear;
          if (FPosition = tpBottom) then
            rec := Rect(TabItem.Left, 1,
                        TabItem.Left + TabItem.Width, FTabHeight)
          else
            rec := Rect(TabItem.Left, Height - FTabHeight,
                        TabItem.Left + TabItem.Width, Height -1);
          if bHasImage then
            Inc(rec.Left, FImages.Width);
          // Calculate the Caption rect area.
          DrawTabCaption(oMemBmp.Canvas, rec, i, True, TabItem.Enabled, True);
          // Draw the Caption.
          DrawTabCaption(oMemBmp.Canvas, rec, i, True, TabItem.Enabled, False);

          if USE_THEMES and Options.EnableXPStyles then
          begin
            // NOTE: This is a HACK :[  Do not know the Transparent color.
            if (FPosition = tpBottom) then
            begin
              // Correct the Tab top-left pixel DOT.
              Pixels[TabItem.Left - iHalfTabMargin, 0] :=
                  Pixels[TabItem.Left - iHalfTabMargin - 1, 0];
              // Correct the Tab top-right pixel DOT.
              Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width, 0] :=
                  Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width + 1, 0];
            end
            else
            begin
              // Correct the Tab top-left pixel DOT.
              Pixels[TabItem.Left - iHalfTabMargin, Height - FTabHeight] :=
                  Pixels[TabItem.Left - iHalfTabMargin - 1, Height - FTabHeight];
              // Correct the Tab top-right pixel DOT.
              Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width, Height - FTabHeight] :=
                  Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width + 1, Height - FTabHeight];
            end;
          end
          else
          begin
            // Draw the left-outer Border.
            Pen.Color := Options.BorderColors.Highlight;
            if (FPosition = tpBottom) then
            begin
              MoveTo(TabItem.Left - iHalfTabMargin, 0);
              LineTo(TabItem.Left - iHalfTabMargin, FTabHeight - 1);
            end
            else
            begin
              MoveTo(TabItem.Left - iHalfTabMargin, Height - FTabHeight + 2);
              LineTo(TabItem.Left - iHalfTabMargin, Height + 1);
              // Correct the Button border DOT.
              Pixels[TabItem.Left - iHalfTabMargin, Height - FTabHeight + 1] :=
                  Pixels[TabItem.Left - iHalfTabMargin - 1, Height - FTabHeight + 1];
              //
              Pixels[TabItem.Left - iHalfTabMargin + 1, Height - FTabHeight + 2] :=
                  Pixels[TabItem.Left - iHalfTabMargin, Height - FTabHeight + 2];
            end;

            // Draw the bottom-outer or top-outer, and the right-outter Border.
            Pen.Color := Options.BorderColors.ShadowDark;
            if (FPosition = tpBottom) then
            begin
              // right-outter
              MoveTo(TabItem.Left + iHalfTabMargin + TabItem.Width, 0);
              LineTo(TabItem.Left + iHalfTabMargin + TabItem.Width, FTabHeight - 1);
              // bottom-outer
              MoveTo(TabItem.Left - iHalfTabMargin + 1, FTabHeight - 1);
              LineTo(TabItem.Left + iHalfTabMargin + TabItem.Width, FTabHeight - 1);
            end
            else
            begin
              // right-outter
              MoveTo(TabItem.Left + iHalfTabMargin + TabItem.Width, Height - FTabHeight + 2);
              LineTo(TabItem.Left + iHalfTabMargin + TabItem.Width, Height);
              // Correct the Button border DOT.
              Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width, Height - FTabHeight + 1] :=
                  Pixels[TabItem.Left + iHalfTabMargin + TabItem.Width + 1, Height - FTabHeight + 1];
              // top-outter
              Pen.Color := Options.BorderColors.Highlight;
              MoveTo(TabItem.Left - iHalfTabMargin + 1, Height - FTabHeight + 1);
              LineTo(TabItem.Left + iHalfTabMargin + TabItem.Width, Height - FTabHeight + 1);
            end;
          end;
        end
        else
        ////////////////////////////////////////////////////////////////////////
        // Draw the Unselected Tab's ///////////////////////////////////////////
        ////////////////////////////////////////////////////////////////////////
        begin
          Brush.Style := bsClear;

          // Draw the tab Image.
          if (FPosition = tpBottom) then
            rec := Rect(TabItem.Left, 1,
                        TabItem.Left + TabItem.Width, FTabHeight)
          else
            rec := Rect(TabItem.Left, Height - FTabHeight + 3,
                        TabItem.Left + TabItem.Width, Height);
          bHasImage := DrawTabImage(oMemBmp.Canvas, rec, i, False, True);

          // Draw the tab Caption.
          Brush.Style := bsClear;
          if (FPosition = tpBottom) then
            rec := Rect(TabItem.Left, 2,
                        TabItem.Left + TabItem.Width, FTabHeight)
          else
            rec := Rect(TabItem.Left, Height - FTabHeight + 2,
                        TabItem.Left + TabItem.Width, Height);
          if bHasImage then
            Inc(rec.Left, FImages.Width);
          // Calculate the Caption rect area.
          DrawTabCaption(oMemBmp.Canvas, rec, i, False, TabItem.Enabled, True);
          // Draw the Caption.          
          DrawTabCaption(oMemBmp.Canvas, rec, i, False, TabItem.Enabled, False);

          // NOTE: The oSepBmp is used to simulate a TBX and TB2000 Separator
          //       object. We need rec.Left to equal zero, because this is
          //       currently the way TBX and TB2000 draws the Separator's.

          // Draw the left-side Separator.
          if (bIsFirst and not bIsFirstTab) or bIsFirstTab then
          begin
            // Copy the control Background onto the Separator.
            recSep := Rect(0, 0, oSepBmp.Width, oSepBmp.Height);
            rec := Rect(TabItem.Left - iHalfTabMargin - 1, 0,
                          TabItem.Left - iHalfTabMargin + 2, Height);
            oSepBmp.Canvas.CopyRect(recSep, oMemBmp.Canvas, rec);
            // Draw the Separator.
            if (FPosition = tpBottom) then
              Dec(recSep.Bottom, (Height-FTabHeight))
            else
              Inc(recSep.Top, (Height-FTabHeight));
            DrawTabSeparator(oSepBmp.Canvas, recSep);
            // Copy the Separator onto the control Background.
            oMemBmp.Canvas.CopyRect(rec, oSepBmp.Canvas, recSep);
          end;

          // Draw the right-side Separator's.
          recSep := Rect(0, 0, oSepBmp.Width, oSepBmp.Height);
          rec := Rect(TabItem.Left + TabItem.Width + iHalfTabMargin, 0,
                      TabItem.Left + TabItem.Width + iHalfTabMargin + 4, Height);
          oSepBmp.Canvas.CopyRect(recSep, oMemBmp.Canvas, rec);
          if (FPosition = tpBottom) then
            Dec(recSep.Bottom, (Height-FTabHeight))
          else
            Inc(recSep.Top, (Height-FTabHeight));
          DrawTabSeparator(oSepBmp.Canvas, recSep);
          oMemBmp.Canvas.CopyRect(rec, oSepBmp.Canvas, recSep);
        end;
      end;
    end;

    // Draw the Bitmap onto the screen.
    BitBlt(Canvas.Handle, 0, 0, oMemBmp.Width, oMemBmp.Height,
           oMemBmp.Canvas.Handle, 0, 0, SRCCOPY);
  finally
    oMemBmp.Free;
    oSepBmp.Free;
  end;
end;

procedure TTBXImageTabSet.ScrollButtonClick(Sender: TObject; Button: TUDBtnType);
begin
  if (csDesigning in ComponentState) then Exit;

  FirstIndex := TTabScroller(Sender).Position;

  if not (csLoading in ComponentState) then
    if Assigned(FOnScrollClick) then
      FOnScrollClick(Self, Button);

  if Options.ChangeOnScroll then
  begin
    if (Button = btNext) then
      TabIndex := TabIndex + 1
    else
      TabIndex := TabIndex - 1;
  end;
end;

procedure TTBXImageTabSet.UpdateEffectiveColor;
begin
  if (Color = clNone) then
  begin
    if (Parent <> nil) then
    begin
      if (Parent is TTBDock) then
        FEffectiveColor := GetEffectiveColor(Parent)
      else
        FEffectiveColor := CurrentTheme.GetViewColor(VT_TOOLBAR);
    end
    else
      FEffectiveColor := CurrentTheme.GetViewColor(VT_TOOLBAR);
  end
  else
    FEffectiveColor := Color;
end;

procedure TTBXImageTabSet.CMColorChanged(var Message: TMessage);
begin
  UpdateEffectiveColor;
  Invalidate;
end;

procedure TTBXImageTabSet.CMDialogChar(var Message: TCMDialogChar);
var
  i: Integer;
begin
  for i := 0 to Items.Count-1 do
  begin
    if IsAccel(Message.CharCode, Items[i].Caption) then
    begin
      Message.Result := 1;
      SetTabIndex(i);
      Break;
    end;
  end;
  inherited;
end;

procedure TTBXImageTabSet.CMFontChanged(var Message: TMessage);
begin
  inherited;
  Canvas.Font := Font;
  Invalidate;
end;

procedure TTBXImageTabSet.CMParentColorChanged(var Message: TMessage);
begin
  if (Color = clNone) then
  begin
    UpdateEffectiveColor;
    if (Message.WParam = 0) then
    begin
      Message.WParam := 1;
      Message.LParam := EffectiveColor;
    end;
  end;
  inherited;
  Invalidate;
end;

procedure TTBXImageTabSet.TBMGetEffectiveColor(var Message: TMessage);
begin
  Message.WParam := EffectiveColor;
  Message.Result := 1;
end;

procedure TTBXImageTabSet.TBMThemeChange(var Message: TMessage);
begin
  case Message.WParam of
    TSC_AFTERVIEWCHANGE:
      begin
        UpdateEffectiveColor;
        Invalidate;
      end;
  end;
end;

procedure TTBXImageTabSet.WMEraseBkgnd(var Message: TWmEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TTBXImageTabSet.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTALLKEYS;
end;

procedure TTBXImageTabSet.WMLButtonDown(var Message: TWMLButtonDown);
var
  i: Integer;
  iExtra   : Integer;
  iMinLeft : Integer;
  iMaxRight: Integer;
begin
  inherited;

  if (csDesigning in ComponentState) then Exit;

  if (Message.YPos <= FTabHeight) then
  begin
    if (Message.YPos < FTabHeight div 2) then
      iExtra := Options.TabMargin div 3
    else
      iExtra := Options.TabMargin div 2;

    for i := FirstIndex to Items.Count-1 do
    begin
      if (Items[i].Width = -1) or (not Items[i].Enabled) then
        Continue;

      iMinLeft  := Items[i].Left - iExtra;
      iMaxRight := Items[i].Left + Items[i].Width + iExtra;
      if (Message.XPos >= iMinLeft) and (Message.XPos <= iMaxRight) then
      begin
        SetTabIndex(i);
        Break;
      end;
    end;
  end;
end;

procedure TTBXImageTabSet.WMNCHitTest(var Message: TWMNCHitTest);
var
  Result: Integer;
begin
  inherited;
  if ShowHint and (not (csDesigning in ComponentState)) then
  begin
    Result := ItemAtPos(ScreenToClient(Point(Message.XPos, Message.YPos)));
    if (Result <> FMouseOverTab) and (Items.Count > 0) then
    begin
      FMouseOverTab := Result;
      DisplayTabHint(Result);
    end;
  end
  else
    FMouseOverTab := -1;
end;

procedure TTBXImageTabSet.WMSize(var Message: TWMSize);
var
  iNumVisTabs: Integer;
  iLastTabPos: Integer;
  //
  function CalcNumTabs(Start, Stop: Integer; Canvas: TCanvas; First: Integer): Integer;
  var
    iWidth: Integer;
  begin
    Result := First;

    while (Start < Stop) and (Result in [0..Items.Count-1]) do
    begin
      with Canvas do
      begin
        iWidth := TextWidth(Items[Result].Caption);
        MeasureTab(Result, iWidth);
        Inc(Start, iWidth + Options.TabMargin); // Next usable position
        if Start <= Stop then
          Inc(Result);
      end;
    end;
  end;
begin
  if (Message.Height < MINIMUM_CONTROL_HEIGHT) then
    Message.Height := MINIMUM_CONTROL_HEIGHT;

  inherited;

  if (FTabHeight > Height) then
    FTabHeight := Height;

  if (Items.Count > 1) then
  begin
    iLastTabPos  := Width - Options.RightMargin;
    iNumVisTabs  := CalcNumTabs(Options.LeftMargin + Options.TabMargin, iLastTabPos, Canvas, 0);
    if (FTabIndex = Items.Count) or (iNumVisTabs > VisibleTabs) or (iNumVisTabs = Items.Count) then
      FirstIndex := Items.Count - iNumVisTabs;
    FNeedAdjust := True;

    AdjustScrollBtnPos;
  end;
  Invalidate;
end;

procedure TTBXImageTabSet.DefineProperties(Filer: TFiler);
begin
  if (Filer is TReader) then
    inherited DefineProperties(Filer);
  // Hide the TabOrder and TabStop properties.
  Filer.DefineProperty('TabOrder', ReadIntData, nil, False);
  Filer.DefineProperty('TabStop', ReadBoolData, nil, False);
end;

procedure TTBXImageTabSet.ReadIntData(Reader: TReader);
begin
  Reader.ReadInteger;
end;

procedure TTBXImageTabSet.ReadBoolData(Reader: TReader);
begin
  Reader.ReadBoolean;
end;

procedure TTBXImageTabSet.SetFirstIndex(const Value: Integer);
begin
  if (Value in [0..Items.Count-1]) then
  begin
    FFirstIndex := Value;
    Invalidate;
  end;
end;

procedure TTBXImageTabSet.SetImages(Value: TCustomImageList);
begin
  if (FImages <> nil) then
    FImages.UnRegisterChanges(FImageChangeLink);
  FImages := Value;
  if (FImages <> nil) then
  begin
    FImages.RegisterChanges(FImageChangeLink);
    FImages.FreeNotification(Self);
  end;
  Invalidate;
end;

procedure TTBXImageTabSet.SetTabHeight(const Value: Integer);
begin
  if (FTabHeight <> Value) then
  begin
    if (Value > Height) then
      FTabHeight := Height
    else
    if (Value < MINIMUM_CONTROL_HEIGHT) then
      FTabHeight := MINIMUM_CONTROL_HEIGHT
    else
      FTabHeight := Value;
    Invalidate;
  end;
end;

procedure TTBXImageTabSet.SetTabIndex(const Value: Integer);
begin
  if (FTabIndex <> Value) then
  begin
    if (Value = -1) or CanChange(Value) then
    begin
      FTabIndex := Value;
      FScrollBtn.Position := FTabIndex;
      AdjustTabPos;
      if (FTabIndex > FFirstIndex + FVisibleTabs - 1) then
        FFirstIndex := FTabIndex;
      Change;        // Do OnChange(...) event.
      Invalidate;
    end;
  end;
end;

function TTBXImageTabSet.GetTabItems: TTabCollection;
begin
  if (FItems = nil) then
    FItems := TTabCollection.Create(Self);
  Result := FItems;
end;

procedure TTBXImageTabSet.SetTabItems(Value: TTabCollection);
begin
  Items.Assign(Value);
end;

procedure TTBXImageTabSet.SetTabPosition(const Value: TTabPosition);
begin
  if (FPosition <> Value) then
  begin
    FPosition := Value;
    Invalidate;
  end;
end;

end.

