{*******************************************************************************
  Author      : Steve Pinneo (steveyvr@yahoo.ca)
  Updated     : June 10 2003
  Version     : 0.9
  Components  : TTBXThemeEx

 PURPOSE:

  Support for the Windows XP Tab control theme. TBX's current TBXThemes.pas
  file does not have support for the Windows XP Tab control, although the
  current TBX TBXUxThemes.pas file does define the Tab control parts and
  states.

  Instead of modifying the current TBXThemes.pas file, I have created this
  TBXThemesEx.pas unit to be an extension to the TBXThemes.pas unit. I hope
  one day the Windows XP Tab control will be included in TBX's sources. ;)

  If you would prefer to add the Windows XP Tab control theme to the current
  TBX sources, then follow these untested instructions:

    TBXThemes.pas

      1) Search for the 1st instance of "SPIN_THEME", then add this below it:

          TAB_THEME: THandle;

      2) Search for the 2nd instance of "SPIN_THEME", then add this below it:

          Close(TAB_THEME);

      3) Search for the 3rd instance of "SPIN_THEME", then add this below it:

          TAB_THEME := OpenThemeData(FWindowHandle, 'TAB');

      4) Search for "procedure PaintStatusBar", then add this below it:

          procedure PaintTabSheet(Canvas: TCanvas; R: TRect; Part, State: Integer); virtual;

      5) Search for "constructor TTBXTheme", then add this below the constructor:

          procedure TTBXTheme.PaintTabSheet(Canvas: TCanvas; R: TRect; Part, State: Integer);
          begin
            if USE_THEMES then
              DrawThemeBackground(TAB_THEME, Canvas.Handle, Part, State, R, nil)
            else
              FillRectEx(Canvas, R, Canvas.Brush.Color);
          end;

    TBXImageTabSet.pas

      1) Search for "CurrentTheme.PaintTabSheet". Enable it, then disable or
         remove the next line which reads:

          DrawThemeBackground(TAB_THEME, ACanvas.Handle, TABP_TABITEM, TIS_SELECTED, ARect, nil)

      2) Remove "TBXThemesEx" from the implementation uses clause.

    tbxtab_d6.dpk

      1) Remove the file TBXThemesEx.pas (this file) from the contains clause.

    Last but not least, save and recompile the TBX and TTBXImageTabSet packages.

*******************************************************************************}

unit TBXThemesEx;

interface

uses
  Windows, Messages, Classes, Forms, Graphics, ImgList, TBXThemes;

{$I TB2Ver.inc}

type
  TTBXThemeEx = class(TTBXTheme)
  public
    constructor Create(const AName: string); override;
    destructor Destroy; override;
    procedure PaintTabSheet(Canvas: TCanvas; R: TRect; Part, State: Integer); virtual;
  end;

var
  TAB_THEME: THandle;
  
var
  USE_THEMES: Boolean;
  
implementation

uses
  SysUtils, TBXUtils, TBXUxThemes;

type
  TTBXThemeManagerEx = class
  private
    FEnableVisualStyles: Boolean;
    FNeedFreeXPThemes: Boolean;
    FNotifies: TList;
    FWindowHandle: HWND;
    procedure SetEnableVisualStyles(Value: Boolean);
  protected
    procedure VisualStylesClose;
    procedure VisualStylesOpen;
    procedure UpdateVariables;
    procedure WndProc(var Message: TMessage);
  public
    constructor Create;
    destructor Destroy; override;
    function  Broadcast(Msg: Cardinal; Param1, Param2: Integer): Integer;
    procedure Notify;
    procedure AddNotification(AObject: TObject);
    procedure RemoveNotification(AObject: TObject);
    //
    property EnableVisualStyles: Boolean read FEnableVisualStyles write SetEnableVisualStyles;
    property NeedFreeXPThemes: Boolean read FNeedFreeXPThemes write FNeedFreeXPThemes;
  end;

var
  ThemeManagerEx: TTBXThemeManagerEx;

{ TTBXThemeEx }

constructor TTBXThemeEx.Create;
begin
  inherited;
  AddTBXSysChangeNotification(Self);
end;

destructor TTBXThemeEx.Destroy;
begin
  RemoveTBXSysChangeNotification(Self);
  inherited;
end;

procedure TTBXThemeEx.PaintTabSheet(Canvas: TCanvas; R: TRect; Part, State: Integer);
begin
  if USE_THEMES then
    DrawThemeBackground(TAB_THEME, Canvas.Handle, Part, State, R, nil)
  else
    FillRectEx(Canvas.Handle, R, Canvas.Brush.Color);
end;

{ TTBXThemeManagerEx }

procedure TTBXThemeManagerEx.AddNotification(AObject: TObject);
begin
  FNotifies.Add(AObject);
end;

function TTBXThemeManagerEx.Broadcast(Msg: Cardinal; Param1, Param2: Integer): Integer;
var
  I: Integer;
  M: TMessage;
begin
  if (FNotifies.Count > 0) then
  begin
    M.Msg := Msg;
    M.WParam := Param1;
    M.LParam := Param2;
    M.Result := 0;
    for I := 0 to FNotifies.Count-1 do
      TObject(FNotifies[I]).Dispatch(M);
    Result := M.Result;
  end
  else
    Result := 0;
end;

constructor TTBXThemeManagerEx.Create;
begin
  FEnableVisualStyles := True;
  FNeedFreeXPThemes := False;
  FNotifies := TList.Create;
  FWindowHandle := {$IFDEF JR_D6}Classes.{$ENDIF}AllocateHWnd(WndProc);
  UpdateVariables;
end;

destructor TTBXThemeManagerEx.Destroy;
begin
  {$IFDEF JR_D6}Classes.{$ENDIF}DeallocateHWnd(FWindowHandle);
  FNotifies.Free;
  inherited;
end;

procedure TTBXThemeManagerEx.Notify;
begin
  if (FNotifies.Count > 0) then
  begin
    Broadcast(TBX_SYSCOMMAND, TSC_BEFOREVIEWCHANGE, 0);
    Broadcast(TBX_SYSCOMMAND, TSC_VIEWCHANGE, 0);
    Broadcast(TBX_SYSCOMMAND, TSC_AFTERVIEWCHANGE, 0);
  end;
end;

procedure TTBXThemeManagerEx.RemoveNotification(AObject: TObject);
begin
  FNotifies.Remove(AObject);
end;

procedure TTBXThemeManagerEx.SetEnableVisualStyles(Value: Boolean);
begin
  if (FEnableVisualStyles <> Value) then
  begin
    FEnableVisualStyles := Value;
    UpdateVariables;
    Notify;
  end;
end;

procedure TTBXThemeManagerEx.UpdateVariables;
begin
  VisualStylesClose;
  VisualStylesOpen;
end;

procedure TTBXThemeManagerEx.VisualStylesClose;
  //
  procedure Close(var ATheme: THandle);
  begin
    if (ATheme <> 0) then
    begin
      CloseThemeData(ATheme);
      ATheme := 0;
    end;
  end;
begin
  if USE_THEMES then
  begin
    Close(TAB_THEME);
  end;
  if NeedFreeXPThemes then
    FreeXPThemes;
end;

procedure TTBXThemeManagerEx.VisualStylesOpen;
begin
  USE_THEMES := False;
  if (Win32Platform = VER_PLATFORM_WIN32_NT) and ((Win32MajorVersion > 5) or
     ((Win32MajorVersion = 5) and (Win32MinorVersion >= 1))) and EnableVisualStyles then
  begin
    NeedFreeXPThemes := CanUseXPThemes;
    InitXPThemes;
    USE_THEMES := CanUseXPThemes;
    try
      TAB_THEME := OpenThemeData(FWindowHandle, 'TAB');
    except
      VisualStylesClose;
    end;
  end;
end;

procedure TTBXThemeManagerEx.WndProc(var Message: TMessage);
const
  ActiveFlags: array [Boolean] of Integer = (TSC_APPDEACTIVATE, TSC_APPACTIVATE);
begin
  case Message.Msg of
    WM_DISPLAYCHANGE, WM_SYSCOLORCHANGE, WM_THEMECHANGED:
      begin
        UpdateVariables;
        Notify;
      end;
    WM_ACTIVATEAPP:
      Broadcast(TBX_SYSCOMMAND, ActiveFlags[Boolean(Message.WParam)], 0);
  end;
  with Message do
    Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

initialization
  ThemeManagerEx := TTBXThemeManagerEx.Create;

finalization
  ThemeManagerEx.Free;

end.

