unit mbTBXEdit;

{$I mxs.inc}

interface

uses
  SysUtils, Classes, Controls, StdCtrls, Windows, Messages, Graphics, TBX,
  TBXUtils, TBXThemes;

type
  TMouseLoc = (mlNone, mlOver, mlDown);

  TmbTBXEdit = class(TEdit)
  private
   FMouseOver: boolean;
   FThemed: boolean;
   FAlignment: TAlignment;
   FDefaultPasswordSymbol: boolean;
   FHotTrack: boolean;

   procedure SetHotTrack(Value: boolean);
   procedure SetThemed(Value: boolean);
   procedure SetAlignment(Value: TAlignment);
   procedure SetDefaultPasswordSymbol(Value: boolean);
  protected
   FMouseLoc: TMouseLoc;

   procedure WMNCPaint(var Msg: TWMNCPaint); message WM_NCPAINT;
   procedure TBMThemeChange(var Message: TMessage); message TBM_THEMECHANGE;
   procedure Resize; override;
   procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
   procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure CMGotFocus(var Message: TCMGotFocus); message CM_ENTER;
   procedure CMLostFocus(var Message: TCMLostFocus); message CM_EXIT;
   procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
   procedure RPaint;
   procedure CreateParams(var Params: TCreateParams); override;
  public
   constructor Create(AOwner: TComponent); override;
   destructor Destroy; override;
  published
   property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
   property Anchors;
   property AutoSelect;
   property AutoSize;
   property BiDiMode;
   property CharCase;
   property Color;
   property Constraints;
   property DefaultPasswordSymbol: boolean read FDefaultPasswordSymbol write SetDefaultPasswordSymbol default false;
   property HotTracking: boolean read FHotTrack write SetHotTrack default true;
   property DragCursor;
   property DragKind;
   property DragMode;
   property Enabled;
   property Font;
   property HideSelection;
   property ImeMode;
   property ImeName;
   property MaxLength;
   property OEMConvert;
   property ParentBiDiMode;
   property ParentFont;
   property ParentColor default false;
   property ParentShowHint;
   property PasswordChar;
   property PopupMenu;
   property ReadOnly;
   property ShowHint;
   property TabOrder;
   property TabStop;
   property Text;
   property Visible;
   property Themed: boolean read FThemed write SetThemed default true;

   property OnResize;
   property OnChange;
   property OnClick;
   property OnContextPopup;
   property OnDblClick;
   property OnDragDrop;
   property OnDragOver;
   property OnEndDock;
   property OnEndDrag;
   property OnEnter;
   property OnExit;
   property OnKeyDown;
   property OnKeyPress;
   property OnKeyUp;
   property OnMouseDown;
   property OnMouseMove;
   property OnMouseUp;
   property OnStartDock;
   property OnStartDrag;
  end;

implementation

constructor TmbTBXEdit.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 ParentColor := false;
 AddThemeNotification(Self);
 FMouseLoc := mlNone;
 FMouseOver := false;
 DoubleBuffered := true;
 FThemed := true;
 FAlignment := taLeftJustify;
 FDefaultPasswordSymbol := false;
 FHotTrack := true;
end;

destructor TmbTBXEdit.Destroy;
begin
 RemoveThemeNotification(Self);
 inherited Destroy;
end;

procedure TmbTBXEdit.CreateParams(var Params: TCreateParams);
var
 Flags: Cardinal;
begin
 inherited;
 Flags := 0;
 case FAlignment of
  taLeftJustify: Flags := ES_LEFT;
  taRightJustify: Flags := ES_RIGHT;
  taCenter: Flags := ES_CENTER;
 end;

 if FDefaultPasswordSymbol then
  Flags := Flags or ES_PASSWORD
 else
  Flags := Flags or 0;

 Params.Style := Params.Style or Flags;
end;

procedure TmbTBXEdit.TBMThemeChange(var Message: TMessage);
begin
if Message.WParam = TSC_VIEWCHANGE then
 RPaint;
end;

procedure TmbTBXEdit.RPaint;
var
 r: TRect;
begin
 r := ClientRect;
 RedrawWindow(Handle, @R, 0, RDW_INVALIDATE or RDW_UPDATENOW or RDW_FRAME);
end;

procedure SetItemInfo(state: string; var IInfo: TTBXItemInfo; var EInfo: TTBXEditInfo);
begin
   FillChar(IInfo, SizeOf(TTBXItemInfo), 0);
   IInfo.ViewType := TVT_NORMALTOOLBAR;
   IInfo.ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE;
   IInfo.Enabled := True;

   FillChar(EInfo, SizeOf(TTBXEditInfo), 0);
   EInfo.RightBtnWidth := CurrentTheme.GetIntegerMetrics(TMI_EDIT_BTNWIDTH);
   EInfo.RightBtnInfo.ButtonType := EBT_DROPDOWN;

   if state = 'normal' then
   begin
     IInfo.Pushed := false;
     IInfo.Selected := false;
     IInfo.HoverKind := hkNone;
     EInfo.RightBtnInfo.ButtonState := 0;
   end;
   if state = 'over' then
   begin
     IInfo.Pushed := false;
     IInfo.Selected := false;
     IInfo.HoverKind := hkMouseHover;
     EInfo.RightBtnInfo.ButtonState := EBDS_HOT;
   end;
   if state = 'down' then
   begin
     IInfo.Pushed := true;
     IInfo.Selected := false;
     IInfo.HoverKind := hkMouseHover;
     EInfo.RightBtnInfo.ButtonState := EBDS_PRESSED;
   end;
end;

procedure TmbTBXEdit.WMNCPaint(var Msg: TWMNCPaint);
var
 DC: HDC;
 ii: TTBXItemInfo;
 ed: TTBXEditInfo;
 R: TRect;
 C: TCanvas;
begin
 inherited;
 if not FThemed then Exit;
 C := TCanvas.Create;
 DC := GetWindowDC(Handle);
 GetWindowRect(Handle, R);
 OffsetRect(R, -R.Left, -R.Top);
 C.Handle := DC;
 ExcludeClipRect(DC, 2, 2, R.Right - 1, R.Bottom - 1); // Important
 // set info
 case FMouseLoc of
  mlNone: SetItemInfo('normal', ii, ed);
  mlOver:
   if FHotTrack then
    SetItemInfo('down', ii, ed)
   else
    SetItemInfo('normal', ii, ed);
  mlDown: SetItemInfo('down', ii, ed);
 end;
 if Focused then SetItemInfo('down', ii, ed);
 if (csDesigning in ComponentState) or (Enabled = false) then SetItemInfo('normal', ii, ed);
 ii.Enabled := Enabled;
 // paint the border
 CurrentTheme.PaintFrame(c, R, ii);
 CurrentTheme.PaintBackgnd(c, R, R, R, CurrentTheme.GetViewColor(TVT_NORMALTOOLBAR), false, TVT_NORMALTOOLBAR);
 CurrentTheme.PaintDock(c, R, R, DP_TOP);
 if CurrentTheme.PaintDockBackground then
  CurrentTheme.PaintBackgnd(c, R, R, R, CurrentTheme.GetViewColor(TVT_NORMALTOOLBAR), false, TVT_NORMALTOOLBAR)
 else
  CurrentTheme.PaintDock(c, R, R, DP_TOP);
 CurrentTheme.PaintButton(c, R, ii);
 //hide double border
 C.Brush.Color := Color;
 InflateRect(R, -1, -1);
 C.FillRect(R);
 // clean up
 C.Handle := 0;
 ReleaseDC(Handle, DC);
 C.Free;
end;

procedure TmbTBXEdit.Resize;
begin
 RPaint;
 if Assigned(OnResize) then
  OnResize(Self);
end;

procedure TmbTBXEdit.CMMouseEnter(var Message: TMessage);
begin
  if not Focused then
   begin
    FMouseOver := true;
    FMouseLoc := mlOver;
    if FHotTrack then
    RPaint;
   end;
end;

procedure TmbTBXEdit.CMMouseLeave(var Message: TMessage);
begin
 if not Focused then
  begin
   FMouseOver := false;
   FMouseLoc := mlNone;
   if FHotTrack then
   RPaint;
  end;
end;

procedure TmbTBXEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 FMouseLoc := mlDown;
 RPaint;
 if Assigned(OnMouseDown) then
  OnMouseDown(Self, Button, Shift, x, y);
end;

procedure TmbTBXEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 if not Focused then
  begin
   if FMouseOver then
    FMouseLoc := mlOver
   else
    FMouseLoc := mlNone;
   RPaint;
  end;
 if Assigned(OnMouseUp) then
  OnMouseUp(Self, Button, Shift, x, y);
end;

procedure TmbTBXEdit.CMGotFocus(var Message: TCMGotFocus);
begin
 inherited;
 FMouseLoc := mlOver;
 RPaint;
end;

procedure TmbTBXEdit.CMLostFocus(var Message: TCMLostFocus);
begin
 inherited;
 FMouseLoc := mlNone;
 RPaint;
end;

procedure TmbTBXEdit.CMEnabledChanged(var Message: TMessage);
begin
 inherited;
 RPaint;
end;

procedure TmbTBXEdit.SetThemed(Value: boolean);
begin
 if FThemed <> Value then
  begin
   FThemed := Value;
   RPaint;
  end;
end;

procedure TmbTBXEdit.SetAlignment(Value: TAlignment);
begin
 if FAlignment <> Value then
  begin
   FAlignment := Value;
   RecreateWnd;
   Invalidate;
  end;
end;

procedure TmbTBXEdit.SetDefaultPasswordSymbol(Value: boolean);
begin
 if FDefaultPasswordSymbol <> Value then
  begin
   FDefaultPasswordSymbol := Value;
   RecreateWnd;
   Invalidate;
  end;
end;

procedure TmbTBXEdit.SetHotTrack(Value: boolean);
begin
 if FHotTrack <> Value then
  begin
   FHotTrack := Value;
   Invalidate;
  end;
end;

end.
