unit mbTBXSplitter;

{$I mxs.inc}

interface

uses
  SysUtils, Classes, Controls, StdCtrls, Windows, Messages, Graphics, TBX,
  TBXUtils, TBXThemes, Forms;

type
  TMouseLoc = (mlNone, mlOver, mlDown);
  NaturalNumber = 1..High(Integer);
  TCanResizeEvent = procedure(Sender: TObject; var NewSize: Integer; var Accept: Boolean) of object;

  TmbTBXSplitter = class(TGraphicControl)
  private
   FMouseLoc: TMouseLoc;
   FMouseOver: boolean;
   FMouseDown: boolean;
   FActiveControl: TWinControl;
   FAutoSnap: Boolean;
   FControl: TControl;
   FDownPos: TPoint;
   FMinSize: NaturalNumber;
   FMaxSize: Integer;
   FNewSize: Integer;
   FOldKeyDown: TKeyEvent;
   FOldSize: Integer;
   FSplit: Integer;
   FOnCanResize: TCanResizeEvent;
   FOnMoved: TNotifyEvent;
   FOnPaint: TNotifyEvent;
   FHotTrack: boolean;

   procedure SetHotTrack(Value: boolean);
   procedure CalcSplitSize(X, Y: Integer; var NewSize, Split: Integer);
   function FindControl: TControl;
   procedure FocusKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
   procedure UpdateControlSize;
   procedure UpdateSize(X, Y: Integer);
  protected
   procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
   procedure TBMThemeChange(var Message: TMessage); message TBM_THEMECHANGE;
   procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
   procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
   procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
   procedure SetItemInfo(state: string; var IInfo: TTBXItemInfo);
   function CanResize(var NewSize: Integer): Boolean; reintroduce; virtual;
   function DoCanResize(var NewSize: Integer): Boolean; virtual;
   procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
   procedure Paint; override;
   procedure RequestAlign; override;
   procedure StopSizing; dynamic;
  public
   constructor Create(AOwner: TComponent); override;
   destructor Destroy; override;
   property Canvas;
  published
   property Align default alLeft;
   property AutoSnap: Boolean read FAutoSnap write FAutoSnap default True;
   property HotTracking: boolean read FHotTrack write SetHotTrack default true;
   property Cursor default crHSplit;
   property Constraints;
   property MinSize: NaturalNumber read FMinSize write FMinSize default 30;
   property Visible;
   property Width default 3;
   property OnCanResize: TCanResizeEvent read FOnCanResize write FOnCanResize;
   property OnMoved: TNotifyEvent read FOnMoved write FOnMoved;
   property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
   property OnMouseDown;
   property OnMouseMove;
   property OnMouseUp;
  end;

implementation

type
  TWinControlAccess = class(TWinControl);

constructor TmbTBXSplitter.Create(AOwner: TComponent);
begin
 inherited Create(AOwner);
 AddThemeNotification(Self);
 FMouseLoc := mlNone;
 FMouseOver := false;
 FMouseDown := false;
 FAutoSnap := True;
 Align := alLeft;
 Width := 3;
 Cursor := crHSplit;
 FMinSize := 30;
 FOldSize := -1;
 FHotTrack := true;
end;

destructor TmbTBXSplitter.Destroy;
begin
 RemoveThemeNotification(Self);
 inherited Destroy;
end;

procedure TmbTBXSplitter.TBMThemeChange(var Message: TMessage);
begin
if Message.WParam = TSC_VIEWCHANGE then
 Invalidate;
end;

function TmbTBXSplitter.FindControl: TControl;
var
  P: TPoint;
  I: Integer;
  R: TRect;
begin
  Result := nil;
  P := Point(Left, Top);
  case Align of
    alLeft: Dec(P.X);
    alRight: Inc(P.X, Width);
    alTop: Dec(P.Y);
    alBottom: Inc(P.Y, Height);
  else
    Exit;
  end;
  for I := 0 to Parent.ControlCount - 1 do
  begin
    Result := Parent.Controls[I];
    if Result.Visible and Result.Enabled then
    begin
      R := Result.BoundsRect;
      if (R.Right - R.Left) = 0 then
        if Align in [alTop, alLeft] then
          Dec(R.Left)
        else
          Inc(R.Right);
      if (R.Bottom - R.Top) = 0 then
        if Align in [alTop, alLeft] then
          Dec(R.Top)
        else
          Inc(R.Bottom);
      if PtInRect(R, P) then Exit;
    end;
  end;
  Result := nil;
end;

procedure TmbTBXSplitter.RequestAlign;
begin
  inherited RequestAlign;
  if (Cursor <> crVSplit) and (Cursor <> crHSplit) then Exit;
  if Align in [alBottom, alTop] then
    Cursor := crVSplit
  else
    Cursor := crHSplit;
end;

function TmbTBXSplitter.DoCanResize(var NewSize: Integer): Boolean;
begin
 Result := CanResize(NewSize);
 if Result and (NewSize <= MinSize) and FAutoSnap then
  NewSize := 0;
end;

function TmbTBXSplitter.CanResize(var NewSize: Integer): Boolean;
begin
 Result := True;
 if Assigned(FOnCanResize) then
  FOnCanResize(Self, NewSize, Result);
end;

procedure TmbTBXSplitter.UpdateControlSize;
begin
  if FNewSize <> FOldSize then
  begin
    case Align of
      alLeft: FControl.Width := FNewSize;
      alTop: FControl.Height := FNewSize;
      alRight:
        begin
          Parent.DisableAlign;
          try
            FControl.Left := FControl.Left + (FControl.Width - FNewSize);
            FControl.Width := FNewSize;
          finally
            Parent.EnableAlign;
          end;
        end;
      alBottom:
        begin
          Parent.DisableAlign;
          try
            FControl.Top := FControl.Top + (FControl.Height - FNewSize);
            FControl.Height := FNewSize;
          finally
            Parent.EnableAlign;
          end;
        end;
    end;
    Update;
    if Assigned(FOnMoved) then
     FOnMoved(Self);
    FOldSize := FNewSize;
  end;
end;

procedure TmbTBXSplitter.CalcSplitSize(X, Y: Integer; var NewSize, Split: Integer);
var
  S: Integer;
begin
  if Align in [alLeft, alRight] then
    Split := X - FDownPos.X
  else
    Split := Y - FDownPos.Y;
  S := 0;
  case Align of
    alLeft: S := FControl.Width + Split;
    alRight: S := FControl.Width - Split;
    alTop: S := FControl.Height + Split;
    alBottom: S := FControl.Height - Split;
  end;
  NewSize := S;
  if S < FMinSize then
    NewSize := FMinSize
  else if S > FMaxSize then
    NewSize := FMaxSize;
  if S <> NewSize then
  begin
    if Align in [alRight, alBottom] then
      S := S - NewSize else
      S := NewSize - S;
    Inc(Split, S);
  end;
end;

procedure TmbTBXSplitter.UpdateSize(X, Y: Integer);
begin
 CalcSplitSize(X, Y, FNewSize, FSplit);
end;

procedure TmbTBXSplitter.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  NewSize, Split: Integer;
begin
  inherited;
  if (ssLeft in Shift) and Assigned(FControl) then
  begin
    CalcSplitSize(X, Y, NewSize, Split);
    if DoCanResize(NewSize) then
    begin
      FNewSize := NewSize;
      FSplit := Split;
      UpdateControlSize;
    end;
  end;
end;

procedure TmbTBXSplitter.FocusKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if Key = VK_ESCAPE then
  StopSizing
 else
  if Assigned(FOldKeyDown) then
   FOldKeyDown(Sender, Key, Shift);
end;

procedure TmbTBXSplitter.StopSizing;
begin
 if Assigned(FControl) then
  begin
    FControl := nil;
    if Assigned(FActiveControl) then
     begin
      TWinControlAccess(FActiveControl).OnKeyDown := FOldKeyDown;
      FActiveControl := nil;
     end;
  end;
 if Assigned(FOnMoved) then
  FOnMoved(Self);
end;

procedure TmbTBXSplitter.SetItemInfo(state: string; var IInfo: TTBXItemInfo);
begin
   FillChar(IInfo, SizeOf(TTBXItemInfo), 0);
   IInfo.ViewType := TVT_NORMALTOOLBAR;
   IInfo.ItemOptions := IO_APPACTIVE or IO_TOOLBARSTYLE;
   IInfo.Enabled := Enabled;

   if state = 'normal' then
   begin
     IInfo.Pushed := false;
     IInfo.Selected := false;
     IInfo.HoverKind := hkNone;
   end;
   if state = 'over' then
   begin
     IInfo.Pushed := false;
     IInfo.Selected := false;
     IInfo.HoverKind := hkMouseHover;
   end;
   if state = 'down' then
   begin
     IInfo.Pushed := true;
     IInfo.Selected := false;
     IInfo.HoverKind := hkMouseHover;
   end;
end;

procedure TmbTBXSplitter.CMMouseEnter(var Message: TMessage);
begin
  if not FMouseDown then
   begin
    FMouseOver := true;
    FMouseLoc := mlOver;
    if FHotTrack then
    Invalidate;
   end;
end;

procedure TmbTBXSplitter.CMMouseLeave(var Message: TMessage);
begin
 FMouseOver := false;
 if not FMouseDown then
  begin
   FMouseLoc := mlNone;
   if FHotTrack then
   Invalidate;
  end;
end;

procedure TmbTBXSplitter.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
 I: Integer;
begin
 FMouseDown := true;
 FMouseLoc := mlDown;
 inherited MouseDown(Button, Shift, X, Y);
 if Button = mbLeft then
  begin
    FControl := FindControl;
    FDownPos := Point(X, Y);
    if Assigned(FControl) then
    begin
      if Align in [alLeft, alRight] then
      begin
        FMaxSize := Parent.ClientWidth - FMinSize;
        for I := 0 to Parent.ControlCount - 1 do
          with Parent.Controls[I] do
            if Visible and (Align in [alLeft, alRight]) then Dec(FMaxSize, Width);
        Inc(FMaxSize, FControl.Width);
      end
      else
      begin
        FMaxSize := Parent.ClientHeight - FMinSize;
        for I := 0 to Parent.ControlCount - 1 do
          with Parent.Controls[I] do
            if Align in [alTop, alBottom] then Dec(FMaxSize, Height);
        Inc(FMaxSize, FControl.Height);
      end;
      UpdateSize(X, Y);
      with ValidParentForm(Self) do
        if ActiveControl <> nil then
        begin
          FActiveControl := ActiveControl;
          FOldKeyDown := TWinControlAccess(FActiveControl).OnKeyDown;
          TWinControlAccess(FActiveControl).OnKeyDown := FocusKeyDown;
        end;
    end;
  end;
 Invalidate;
end;

procedure TmbTBXSplitter.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 FMouseDown := false;
 if FMouseOver then
  FMouseLoc := mlOver
 else
  FMouseLoc := mlNone;
 Invalidate;
 inherited;
 if Assigned(FControl) then
  StopSizing;
end;

procedure TmbTBXSplitter.Paint;
var
 R: TRect;
 ii: TTBXItemInfo;
begin
 R := ClientRect;
 Canvas.Brush.Color := Color;
 Canvas.Brush.Style := bsSolid;
 Canvas.FillRect(R);
 case FMouseLoc of
  mlNone: SetItemInfo('normal', ii);
  mlOver:
   if FHotTrack then
    SetItemInfo('over', ii)
   else
    SetItemInfo('normal', ii);
  mlDown: SetItemInfo('down', ii);
 end;
 if (csDesigning in ComponentState) then SetItemInfo('normal', ii);
 ii.Enabled := Enabled;
 Canvas.Pen.Color := Blend(CurrentTheme.GetItemColor(ii), clBtnShadow, 50);
 Canvas.Brush.Color := CurrentTheme.GetItemColor(ii);
 Canvas.Rectangle(R);
 //uncomment this is you want the dotted frame drawn at designtime
{ if csDesigning in ComponentState then
  with Canvas do
   begin
    ExcludeClipRect(Canvas.Handle, 1, 1, Width - 1, Height - 1);
    FMouseLoc := mlNone;
    Pen.Style := psDot;
    Pen.Mode := pmXor;
    Pen.Color := $00FFD8CE;
    Brush.Style := bsClear;
    Rectangle(ClientRect);
  end;}
 if Assigned(FOnPaint) then
  FOnPaint(Self);
end;

procedure TmbTBXSplitter.CMEnabledChanged(var Message: TMessage);
begin
 inherited;
 Invalidate;
end;

procedure TmbTBXSplitter.SetHotTrack(Value: boolean);
begin
 if FHotTrack <> Value then
  begin
   FHotTrack := Value;
   Invalidate;
  end;
end;

end.
